@@ -1,12 +1,9 @@
 /*
 ** LuaLDAP
-** See Copyright Notice in license.html
-** $Id: lualdap.c,v 1.48 2007-12-14 15:11:22 carregal Exp $
+** See Copyright Notice in license.md
 */
 
-#include <stdlib.h>
 #include <string.h>
-#include <assert.h>
 
 #ifdef WIN32
 #include <Winsock2.h>
@@ -17,13 +14,22 @@
 #ifdef WINLDAP
 #include "open2winldap.h"
 #else
-#include "ldap.h"
+#include <ldap.h>
 #endif
 
 #include <lua.h>
 #include <lauxlib.h>
 
-#include "compat-5.3.h"
+#if LUA_VERSION_NUM < 502
+#define lua_rawlen(L,i)		lua_objlen(L, (i))
+#define luaL_setmetatable(L,n)	\
+	(luaL_getmetatable(L, n), lua_setmetatable(L, -2))
+#define luaL_newlibtable(L,l)	\
+	lua_createtable(L, 0, sizeof(l)/sizeof((l)[0]) - 1)
+#define luaL_newlib(L,l)	\
+	(luaL_newlibtable(L, (l)), luaL_setfuncs(L, (l), 0))
+#define luaL_setfuncs(L,l,n)		luaL_register(L, NULL, (l))
+#endif
 
 
 #ifdef WINLDAPAPI
@@ -105,7 +111,6 @@ static int faildirect (lua_State *L, const char *errmsg) {
 */
 static conn_data *getconnection (lua_State *L) {
 	conn_data *conn = (conn_data *)luaL_checkudata (L, 1, LUALDAP_CONNECTION_METATABLE);
-	luaL_argcheck(L, conn!=NULL, 1, LUALDAP_PREFIX"LDAP connection expected");
 	luaL_argcheck(L, conn->ld, 1, LUALDAP_PREFIX"LDAP connection is closed");
 	return conn;
 }
@@ -122,15 +127,6 @@ static search_data *getsearch (lua_State *L) {
 }
 
 
-/*
-** Set metatable of userdata on top of the stack.
-*/
-static void lualdap_setmeta (lua_State *L, const char *name) {
-	luaL_getmetatable (L, name);
-	lua_setmetatable (L, -2);
-}
-
-
 /*
 ** Error on option.
 */
@@ -139,21 +135,12 @@ static int option_error (lua_State *L, const char *name, const char *type) {
 }
 
 
-/*
-** Get the field called name of the table at position 2.
-*/
-static void strgettable (lua_State *L, int idx, const char *name) {
-	lua_pushstring (L, name);
-	lua_gettable (L, idx);
-}
-
-
 /*
 ** Get the field named name as a string.
 ** The table MUST be at position 2.
 */
-static const char *strtabparam (lua_State *L, int idx, const char *name, char *def) {
-	strgettable (L, idx, name);
+static const char *strtabparam (lua_State *L, const char *name, char *def) {
+	lua_getfield(L, 2, name);
 	if (lua_isnil (L, -1))
 		return def;
 	else if (lua_isstring (L, -1))
@@ -169,12 +156,12 @@ static const char *strtabparam (lua_State *L, int idx, const char *name, char *d
 ** Get the field named name as an integer.
 ** The table MUST be at position 2.
 */
-static long longtabparam (lua_State *L, int idx, const char *name, int def) {
-	strgettable (L, idx, name);
+static long longtabparam (lua_State *L, const char *name, int def) {
+	lua_getfield(L, 2, name);
 	if (lua_isnil (L, -1))
 		return def;
 	else if (lua_isnumber (L, -1))
-		return (long)lua_tonumber (L, -1);
+		return (long)lua_tointeger (L, -1);
 	else
 		return option_error (L, name, "number");
 }
@@ -184,8 +171,8 @@ static long longtabparam (lua_State *L, int idx, const char *name, int def) {
 ** Get the field named name as a double.
 ** The table MUST be at position 2.
 */
-static double numbertabparam (lua_State *L, int idx, const char *name, double def) {
-	strgettable (L, idx, name);
+static double numbertabparam (lua_State *L, const char *name, double def) {
+	lua_getfield(L, 2, name);
 	if (lua_isnil (L, -1))
 		return def;
 	else if (lua_isnumber (L, -1))
@@ -200,7 +187,7 @@ static double numbertabparam (lua_State *L, int idx, const char *name, double de
 ** The table MUST be at position 2.
 */
 static int booltabparam (lua_State *L, const char *name, int def) {
-	strgettable (L, 2, name);
+	lua_getfield(L, 2, name);
 	if (lua_isnil (L, -1))
 		return def;
 	else if (lua_isboolean (L, -1))
@@ -236,6 +223,7 @@ static void A_init (attrs_data *attrs) {
 ** Increment the bvals counter.
 */
 static BerValue *A_setbval (lua_State *L, attrs_data *a, const char *n) {
+	size_t len;
 	BerValue *ret = &(a->bvals[a->bi]);
 	if (a->bi >= LUALDAP_MAX_VALUES) {
 		luaL_error (L, LUALDAP_PREFIX"too many values");
@@ -244,8 +232,8 @@ static BerValue *A_setbval (lua_State *L, attrs_data *a, const char *n) {
 		value_error (L, n);
 		return NULL;
 	}
-	a->bvals[a->bi].bv_len = lua_rawlen (L, -1);
-	a->bvals[a->bi].bv_val = (char *)lua_tostring (L, -1);
+	a->bvals[a->bi].bv_val = (char *)lua_tolstring (L, -1, &len);
+	a->bvals[a->bi].bv_len = len;
 	a->bi++;
 	return ret;
 }
@@ -357,48 +345,6 @@ static void A_lastattr (lua_State *L, attrs_data *a) {
 }
 
 
-/*
-** Copy a string or a table of strings from Lua to a NULL-terminated array
-** of C-strings.
-*/
-static int table2strarray (lua_State *L, int tab, char *array[], int limit) {
-	if (lua_isstring (L, tab)) {
-		if (limit < 2)
-			return luaL_error (L, LUALDAP_PREFIX"too many arguments");
-		array[0] = (char *)lua_tostring (L, tab);
-		array[1] = NULL;
-	} else if (lua_istable (L, tab)) {
-		int i;
-		int n = lua_rawlen (L, tab);
-		if (limit < (n+1))
-			return luaL_error (L, LUALDAP_PREFIX"too many arguments");
-		for (i = 0; i < n; i++) {
-			lua_rawgeti (L, tab, i+1); /* push table element */
-			if (lua_isstring (L, -1))
-				array[i] = (char *)lua_tostring (L, -1);
-			else {
-				return luaL_error (L, LUALDAP_PREFIX"invalid value #%d", i+1);
-			}
-		}
-		array[n] = NULL;
-	} else 
-		return luaL_error (L, LUALDAP_PREFIX"bad argument #%d (table or string expected, got %s)", tab, lua_typename (L, lua_type (L, tab)));
-	return 0;
-}
-
-
-/*
-** Fill in the struct timeval, according to the timeout parameter.
-*/
-static struct timeval *get_timeout_param (lua_State *L, int idx, struct timeval *st) {
-	double t = numbertabparam (L, idx, "timeout", -1);
-	if(t < 0)
-		return NULL; /* No timeout, block */
-	st->tv_sec = (long)t;
-	st->tv_usec = (long)(1000000 * (t - st->tv_sec));
-	return st;
-}
-
 /*
 ** Get the result message of an operation.
 ** #1 upvalue == connection
@@ -406,7 +352,7 @@ static struct timeval *get_timeout_param (lua_State *L, int idx, struct timeval
 ** #3 upvalue == result code of the message (ADD, DEL etc.) to be received.
 */
 static int result_message (lua_State *L) {
-	struct timeval timeout;
+	struct timeval *timeout = NULL; /* ??? function parameter ??? */
 	LDAPMessage *res;
 	int rc;
 	conn_data *conn = (conn_data *)lua_touserdata (L, lua_upvalueindex (1));
@@ -414,7 +360,7 @@ static int result_message (lua_State *L) {
 	/*int res_code = (int)lua_tonumber (L, lua_upvalueindex (3));*/
 
 	luaL_argcheck (L, conn->ld, 1, LUALDAP_PREFIX"LDAP connection is closed");
-	rc = ldap_result (conn->ld, msgid, LDAP_MSG_ONE, get_timeout_param (L, 1, &timeout), &res);
+	rc = ldap_result (conn->ld, msgid, LDAP_MSG_ONE, timeout, &res);
 	if (rc == 0)
 		return faildirect (L, LUALDAP_PREFIX"result timeout expired");
 	else if (rc < 0) {
@@ -474,7 +420,6 @@ static int create_future (lua_State *L, ldap_int_t rc, int conn, ldap_int_t msgi
 */
 static int lualdap_close (lua_State *L) {
 	conn_data *conn = (conn_data *)luaL_checkudata (L, 1, LUALDAP_CONNECTION_METATABLE);
-	luaL_argcheck(L, conn!=NULL, 1, LUALDAP_PREFIX"LDAP connection expected");
 	if (conn->ld == NULL) /* already closed */
 		return 0;
 #if defined(LDAP_API_FEATURE_X_OPENLDAP) && LDAP_API_FEATURE_X_OPENLDAP >= 20300
@@ -493,7 +438,7 @@ static int lualdap_close (lua_State *L) {
 ** @param #1 LDAP connection.
 ** @param #2 String with username.
 ** @param #3 String with password.
-** @return Boolean.
+** @return LDAP connection on success.
 */
 static int lualdap_bind_simple (lua_State *L) {
 	conn_data *conn = getconnection (L);
@@ -501,20 +446,17 @@ static int lualdap_bind_simple (lua_State *L) {
 	const char *password = luaL_checkstring (L, 3);
 	int err;
 #if defined(LDAP_API_FEATURE_X_OPENLDAP) && LDAP_API_FEATURE_X_OPENLDAP >= 20300
-	struct berval cred = { 0, NULL };
-	cred.bv_len = strlen(password);
-	cred.bv_val = malloc(cred.bv_len+1);
-	strcpy(cred.bv_val, password);
-	err = ldap_sasl_bind_s (conn->ld, who, LDAP_SASL_SIMPLE, &cred, NULL, NULL, NULL);
-	free(cred.bv_val);
-	memset(&cred, 0, sizeof(cred));
+	struct berval *cred = ber_bvstrdup(password);
+	/* FIXME: According to `man 3 ldap_sasl_bind_s` we should pass `NULL` instead of `who`. */
+	err = ldap_sasl_bind_s (conn->ld, who, LDAP_SASL_SIMPLE, cred, NULL, NULL, NULL);
+	ber_bvfree(cred);
 #else
-	err = ldap_simple_bind_s (conn->ld, who, password);
+	err = ldap_bind_s (conn->ld, who, password, LDAP_AUTH_SIMPLE);
 #endif
 	if (err != LDAP_SUCCESS)
 		return faildirect (L, ldap_err2string (err));
-	
-	lua_pushboolean (L, 1);
+
+	lua_pushvalue (L, 1);
 	return 1;
 }
 
@@ -554,8 +496,9 @@ static int lualdap_compare (lua_State *L) {
 	ldap_pchar_t attr = (ldap_pchar_t) luaL_checkstring (L, 3);
 	BerValue bvalue;
 	ldap_int_t rc, msgid;
-	bvalue.bv_val = (char *)luaL_checkstring (L, 4);
-	bvalue.bv_len = lua_rawlen (L, 4);
+	size_t len;
+	bvalue.bv_val = (char *)luaL_checklstring (L, 4, &len);
+	bvalue.bv_len = len;
 	rc = ldap_compare_ext (conn->ld, dn, attr, &bvalue, NULL, NULL, &msgid);
 	return create_future (L, rc, 1, msgid, LDAP_RES_COMPARE);
 }
@@ -717,14 +660,15 @@ static void search_close (lua_State *L, search_data *search) {
 static int next_message (lua_State *L) {
 	search_data *search = getsearch (L);
 	conn_data *conn;
-	struct timeval timeout;
+	struct timeval *timeout = NULL; /* ??? function parameter ??? */
 	LDAPMessage *res;
 	int rc;
 	int ret;
 
 	lua_rawgeti (L, LUA_REGISTRYINDEX, search->conn);
 	conn = (conn_data *)lua_touserdata (L, -1); /* get connection */
-	rc = ldap_result (conn->ld, search->msgid, LDAP_MSG_ONE, get_timeout_param (L, 1, &timeout), &res);
+
+	rc = ldap_result (conn->ld, search->msgid, LDAP_MSG_ONE, timeout, &res);
 	if (rc == 0)
 		return faildirect (L, LUALDAP_PREFIX"result timeout expired");
 	else if (rc == -1)
@@ -793,7 +737,6 @@ static int string2scope (lua_State *L, const char *s) {
 */
 static int lualdap_search_close (lua_State *L) {
 	search_data *search = (search_data *)luaL_checkudata (L, 1, LUALDAP_SEARCH_METATABLE);
-	luaL_argcheck (L, search!=NULL, 1, LUALDAP_PREFIX"LDAP search expected");
 	if (search->conn == LUA_NOREF)
 		return 0;
 	search_close (L, search);
@@ -807,7 +750,7 @@ static int lualdap_search_close (lua_State *L) {
 */
 static void create_search (lua_State *L, int conn_index, int msgid) {
 	search_data *search = (search_data *)lua_newuserdata (L, sizeof (search_data));
-	lualdap_setmeta (L, LUALDAP_SEARCH_METATABLE);
+	luaL_setmetatable (L, LUALDAP_SEARCH_METATABLE);
 	search->conn = LUA_NOREF;
 	search->msgid = msgid;
 	lua_pushvalue (L, conn_index);
@@ -818,18 +761,42 @@ static void create_search (lua_State *L, int conn_index, int msgid) {
 /*
 ** Fill in the attrs array, according to the attrs parameter.
 */
-static int get_attrs_param (lua_State *L, char *attrs[]) {
-	lua_pushstring (L, "attrs");
-	lua_gettable (L, 2);
+static void get_attrs_param (lua_State *L, char *attrs[]) {
+	lua_getfield(L, 2, "attrs");
 	if (lua_isstring (L, -1)) {
 		attrs[0] = (char *)lua_tostring (L, -1);
 		attrs[1] = NULL;
 	} else if (!lua_istable (L, -1))
 		attrs[0] = NULL;
-	else
-		if (table2strarray (L, lua_gettop (L), attrs, LUALDAP_MAX_ATTRS))
-			return 0;
-	return 1;
+	else {
+		int i;
+		int n = lua_rawlen(L, -1);
+		if (LUALDAP_MAX_ATTRS < (n+1))
+			luaL_error (L, LUALDAP_PREFIX"too many arguments");
+		luaL_checkstack(L, n, NULL);
+		for (i = 0; i < n; i++) {
+			lua_rawgeti (L, -1, i+1); /* push table element */
+			if (lua_isstring (L, -1))
+				attrs[i] = (char *)lua_tostring (L, -1);
+			else {
+				luaL_error (L, LUALDAP_PREFIX"invalid value #%d", i+1);
+			}
+		}
+		attrs[n] = NULL;
+	}
+}
+
+
+/*
+** Fill in the struct timeval, according to the timeout parameter.
+*/
+static struct timeval *get_timeout_param (lua_State *L, struct timeval *st) {
+	double t = numbertabparam (L, "timeout", 0.0);
+	if(t <= 0.0)
+		return NULL; /* No timeout, block */
+	st->tv_sec = (long)t;
+	st->tv_usec = (long)(1000000.0 * (t - (double)st->tv_sec));
+	return st;
 }
 
 
@@ -850,15 +817,14 @@ static int lualdap_search (lua_State *L) {
 
 	if (!lua_istable (L, 2))
 		return luaL_error (L, LUALDAP_PREFIX"no search specification");
-	if (!get_attrs_param (L, attrs))
-		return 2;
+	get_attrs_param (L, attrs);
 	/* get other parameters */
 	attrsonly = booltabparam (L, "attrsonly", 0);
-	base = (ldap_pchar_t) strtabparam (L, 2, "base", NULL);
-	filter = (ldap_pchar_t) strtabparam (L, 2, "filter", NULL);
-	scope = string2scope (L, strtabparam (L, 2, "scope", NULL));
-	sizelimit = longtabparam (L, 2, "sizelimit", LDAP_NO_LIMIT);
-	timeout = get_timeout_param (L, 2, &st);
+	base = (ldap_pchar_t) strtabparam (L, "base", NULL);
+	filter = (ldap_pchar_t) strtabparam (L, "filter", NULL);
+	scope = string2scope (L, strtabparam (L, "scope", NULL));
+	sizelimit = longtabparam (L, "sizelimit", LDAP_NO_LIMIT);
+	timeout = get_timeout_param (L, &st);
 
 	rc = ldap_search_ext (conn->ld, base, scope, filter, attrs, attrsonly,
 		NULL, NULL, timeout, sizelimit, &msgid);
@@ -877,13 +843,11 @@ static int lualdap_search (lua_State *L) {
 ** This function is used by `tostring'.
 */
 static int lualdap_conn_tostring (lua_State *L) {
-	char buff[100];
-	conn_data *conn = (conn_data *)lua_touserdata (L, 1);
+	conn_data *conn = luaL_checkudata(L, 1, LUALDAP_CONNECTION_METATABLE);
 	if (conn->ld == NULL)
-		strcpy (buff, "closed");
+		lua_pushfstring (L, "%s (closed)", LUALDAP_CONNECTION_METATABLE);
 	else
-		sprintf (buff, "%p", (void*)conn);
-	lua_pushfstring (L, "%s (%s)", LUALDAP_CONNECTION_METATABLE, buff);
+		lua_pushfstring (L, "%s (%p)", LUALDAP_CONNECTION_METATABLE, (void*)conn);
 	return 1;
 }
 
@@ -893,14 +857,11 @@ static int lualdap_conn_tostring (lua_State *L) {
 ** This function is used by `tostring'.
 */
 static int lualdap_search_tostring (lua_State *L) {
-	char buff[100];
-	search_data *search = (search_data *)lua_touserdata (L, 1);
-	luaL_argcheck (L,search->conn!=LUA_NOREF,1,LUALDAP_PREFIX"LDAP search is closed");
+	search_data *search = luaL_checkudata(L, 1, LUALDAP_SEARCH_METATABLE);
 	if (search->conn == LUA_NOREF)
-		strcpy (buff, "closed");
+		lua_pushfstring (L, "%s (closed)", LUALDAP_SEARCH_METATABLE);
 	else
-		sprintf (buff, "%p", (void*)search);
-	lua_pushfstring (L, "%s (%s)", LUALDAP_SEARCH_METATABLE, buff);
+		lua_pushfstring (L, "%s (%p)", LUALDAP_SEARCH_METATABLE, (void*)search);
 	return 1;
 }
 
@@ -908,8 +869,16 @@ static int lualdap_search_tostring (lua_State *L) {
 /*
 ** Create a metatable.
 */
-static int lualdap_createmeta (lua_State *L) {
-	const luaL_Reg methods[] = {
+static void lualdap_createmeta_conn (lua_State *L) {
+	static const luaL_Reg metamethods[] = {
+		{"__gc", lualdap_close},
+		{"__tostring", lualdap_conn_tostring},
+		/* placeholders */
+		{"__index", NULL},
+		{"__metatable", NULL},
+		{NULL, NULL}
+	};
+	static const luaL_Reg methods[] = {
 		{"close", lualdap_close},
 		{"bind_simple", lualdap_bind_simple},
 		{"add", lualdap_add},
@@ -921,48 +890,43 @@ static int lualdap_createmeta (lua_State *L) {
 		{NULL, NULL}
 	};
 
-	if (!luaL_newmetatable (L, LUALDAP_CONNECTION_METATABLE))
-		return 0;
+	luaL_newmetatable (L, LUALDAP_CONNECTION_METATABLE);
+	luaL_setfuncs(L, metamethods, 0);  /* add metamethods to new metatable */
 
-	/* define methods */
-	luaL_setfuncs(L, methods, 0);
+	luaL_newlibtable(L, methods);  /* create method table */
+	luaL_setfuncs(L, methods, 0);  /* add file methods to method table */
+	lua_setfield(L, -2, "__index");  /* metatable.__index = method table */
 
-	/* define metamethods */
-	lua_pushliteral (L, "__gc");
-	lua_pushcfunction (L, lualdap_close);
-	lua_settable (L, -3);
-
-	lua_pushliteral (L, "__index");
-	lua_pushvalue (L, -2);
-	lua_settable (L, -3);
-
-	lua_pushliteral (L, "__tostring");
-	lua_pushcfunction (L, lualdap_conn_tostring);
-	lua_settable (L, -3);
-
-	lua_pushliteral (L, "__metatable");
 	lua_pushliteral(L,LUALDAP_PREFIX"you're not allowed to get this metatable");
-	lua_settable (L, -3);
+	lua_setfield (L, -2, "__metatable");
 
-	if (!luaL_newmetatable (L, LUALDAP_SEARCH_METATABLE))
-		return 0;
-
-	lua_pushliteral (L, "__gc");
-	lua_pushcfunction (L, lualdap_search_close);
-	lua_settable (L, -3);
-
-	lua_pushliteral (L, "__tostring");
-	lua_pushcclosure (L, lualdap_search_tostring, 1);
-	lua_settable (L, -3);
-
-	lua_pushliteral (L, "__metatable");
-	lua_pushliteral(L,LUALDAP_PREFIX"you're not allowed to get this metatable");
-	lua_settable (L, -3);
-
-	return 0;
+	lua_pop(L, 1);  /* pop metatable */
 }
 
 
+/*
+** Create a metatable.
+*/
+static void lualdap_createmeta_search (lua_State *L) {
+	static const luaL_Reg metamethods[] = {
+		{"__gc", lualdap_search_close},
+		{"__tostring", lualdap_search_tostring},
+		/* placeholders */
+		{"__metatable", NULL},
+		{NULL, NULL}
+	};
+
+	luaL_newmetatable (L, LUALDAP_SEARCH_METATABLE);
+	luaL_setfuncs(L, metamethods, 0);  /* add metamethods to new metatable */
+
+	lua_pushliteral(L,LUALDAP_PREFIX"you're not allowed to get this metatable");
+	lua_setfield (L, -2, "__metatable");
+
+	lua_pop(L, 1);  /* pop metatable */
+}
+
+
+#if !defined(WINLDAP)
 /*
 ** Open and initialize a connection to a server (without binding).
 ** @param #1 String with URI.
@@ -975,12 +939,16 @@ static int lualdap_initialize (lua_State *L) {
 	int lev=7;
 
 	/* Initialize */
-	lualdap_setmeta (L, LUALDAP_CONNECTION_METATABLE);
+	luaL_setmetatable (L, LUALDAP_CONNECTION_METATABLE);
 	conn->version = 0;
+#if defined(LDAP_API_FEATURE_X_OPENLDAP) && LDAP_API_FEATURE_X_OPENLDAP >= 20300
 	err = ldap_initialize (&conn->ld, uri);
 	if (err != LDAP_SUCCESS)
 		return faildirect(L, ldap_err2string(err));
-	
+#else
+#	error Unsupported LDAP library version
+#endif
+
 	/* Set protocol version */
 	conn->version = LDAP_VERSION3;
 	if (ldap_set_option (conn->ld, LDAP_OPT_PROTOCOL_VERSION, &conn->version)
@@ -990,39 +958,40 @@ static int lualdap_initialize (lua_State *L) {
 
 	return 1;
 }
+#endif
 
 /*
-** Open and initialize a connection to a server.
+** Open a connection to a server.
 ** @param #1 String with hostname.
-** @param #2 String with username.
-** @param #3 String with password.
-** @param #4 Boolean indicating if TLS must be used.
+** @param #2 Boolean indicating if TLS must be used.
 ** @return #1 Userdata with connection structure.
 */
-static int lualdap_open_simple (lua_State *L) {
+static int lualdap_open (lua_State *L) {
 	ldap_pchar_t host = (ldap_pchar_t) luaL_checkstring (L, 1);
-	ldap_pchar_t who = (ldap_pchar_t) luaL_optstring (L, 2, NULL);
-	const char *password = luaL_optstring (L, 3, "");
-	int use_tls = lua_toboolean (L, 4);
+	int use_tls = lua_toboolean (L, 2);
 	conn_data *conn = (conn_data *)lua_newuserdata (L, sizeof(conn_data));
 #if defined(LDAP_API_FEATURE_X_OPENLDAP) && LDAP_API_FEATURE_X_OPENLDAP >= 20300
-	struct berval *cred = NULL;
-#endif
 	int err;
+#endif
 
 	/* Initialize */
-	lualdap_setmeta (L, LUALDAP_CONNECTION_METATABLE);
+	luaL_setmetatable (L, LUALDAP_CONNECTION_METATABLE);
 	conn->version = 0;
 #if defined(LDAP_API_FEATURE_X_OPENLDAP) && LDAP_API_FEATURE_X_OPENLDAP >= 20300
 	if (strstr(host, "://") != NULL) {
 		err = ldap_initialize(&conn->ld, host);
 	} else {
-		char *host_with_schema = malloc(strlen(host) + 8);
-		strcpy(host_with_schema, "ldap://");
-		strcat(host_with_schema, host);
-		err = ldap_initialize(&conn->ld, host_with_schema);
-		free(host_with_schema);
-		host_with_schema = NULL;
+		lua_getglobal (L, "string");
+		lua_getfield (L, -1, "gsub");
+		if (!lua_isfunction (L, -1))
+			return faildirect (L, LUALDAP_PREFIX"string.gsub broken");
+		lua_pushvalue (L, 1); /* host */
+		lua_pushstring (L, "(%S+)");
+		lua_pushstring (L, "ldap://%1");
+		lua_call (L, 3, 1); /* string.gsub(host, '(%S+)', 'ldap://%1') */
+		/* ldap_initialize handles a whitespace-separated list of hostnames */
+		err = ldap_initialize(&conn->ld, lua_tostring (L, -1));
+		lua_pop(L, 2);
 	}
 	if (err != LDAP_SUCCESS)
 #else
@@ -1041,17 +1010,45 @@ static int lualdap_open_simple (lua_State *L) {
 		if (rc != LDAP_SUCCESS)
 			return faildirect (L, ldap_err2string (rc));
 	}
-	/* Bind to a server */
-#if defined(LDAP_API_FEATURE_X_OPENLDAP) && LDAP_API_FEATURE_X_OPENLDAP >= 20300
-	cred = ber_bvstrdup(password);
-	err = ldap_sasl_bind_s (conn->ld, who, LDAP_SASL_SIMPLE, cred, NULL, NULL, NULL);
-	ber_bvfree(cred);
-#else
-	err = ldap_bind_s (conn->ld, who, password, LDAP_AUTH_SIMPLE);
-#endif
-	if (err != LDAP_SUCCESS)
-		return faildirect (L, ldap_err2string (err));
+	return 1;
+}
 
+/*
+** Open and initialize a connection to a server.
+** @param #1 String with hostname.
+** @param #2 String with username.
+** @param #3 String with password.
+** @param #4 Boolean indicating if TLS must be used.
+** @return #1 Userdata with connection structure.
+*/
+static int lualdap_open_simple (lua_State *L) {
+	ldap_pchar_t host = (ldap_pchar_t) luaL_checkstring (L, 1);
+	ldap_pchar_t who = (ldap_pchar_t) luaL_optstring (L, 2, "");
+	const char *password = luaL_optstring (L, 3, "");
+	int use_tls = lua_toboolean (L, 4);
+
+	/* Open */
+	lua_pushcfunction (L, lualdap_open);
+	lua_pushstring (L, host);
+	lua_pushboolean (L, use_tls);
+	lua_call (L, 2, 2);
+	if (lua_isnil (L, -2)) {
+		return 2; /* nil, msg */
+	} else {
+		lua_pop (L, 1); /* keep only the userdata (LDAP connection) */
+	}
+
+	/* Bind to a server */
+	lua_pushcfunction (L, lualdap_bind_simple);
+	lua_pushvalue (L, -2); /* connection */
+	lua_pushstring (L, who);
+	lua_pushstring (L, password);
+	lua_call (L, 3, 2);
+	if (lua_isnil (L, -2)) {
+		return 2; /* Pass through in case of errors */
+	} else {
+		lua_pop (L, 2); /* Discard return values from lualdap_bind_simple */
+	}
 	return 1;
 }
 
@@ -1060,15 +1057,12 @@ static int lualdap_open_simple (lua_State *L) {
 ** Assumes the table is on top of the stack.
 */
 static void set_info (lua_State *L) {
-	lua_pushliteral (L, "_COPYRIGHT");
 	lua_pushliteral (L, "Copyright (C) 2003-2007 Kepler Project");
-	lua_settable (L, -3);
-	lua_pushliteral (L, "_DESCRIPTION");
+	lua_setfield (L, -2, "_COPYRIGHT");
 	lua_pushliteral (L, "LuaLDAP is a simple interface from Lua to an LDAP client");
-	lua_settable (L, -3);
-	lua_pushliteral (L, "_VERSION");
-	lua_pushliteral (L, PACKAGE_STRING);
-	lua_settable (L, -3);
+	lua_setfield (L, -2, "_DESCRIPTION");
+	lua_pushliteral (L, "LuaLDAP 1.3.0");
+	lua_setfield (L, -2, "_VERSION");
 }
 
 
@@ -1076,20 +1070,28 @@ static void set_info (lua_State *L) {
 ** Create ldap table and register the open method.
 */
 int luaopen_lualdap (lua_State *L) {
-	struct luaL_Reg lualdap[] = {
+	static const struct luaL_Reg lualdap[] = {
+#if !defined(WINLDAP)
 		{"initialize", lualdap_initialize},
+#endif
+		{"open", lualdap_open},
 		{"open_simple", lualdap_open_simple},
+		/* placeholders */
+		{"_COPYRIGHT", NULL},
+		{"_DESCRIPTION", NULL},
+		{"_VERSION", NULL},
 		{NULL, NULL},
 	};
 
-	lualdap_createmeta (L);
+	lualdap_createmeta_conn (L);
+	lualdap_createmeta_search (L);
 	luaL_newlib(L, lualdap);
-	lua_pushvalue(L, -1);
 /*
    In Lua 5.2 "modules are not expected to set global variables":
    https://www.lua.org/manual/5.2/manual.html#8.2
 */
 #if LUA_VERSION_NUM < 502
+	lua_pushvalue(L, -1);
 	lua_setglobal(L, LUALDAP_TABLENAME);
 #endif
 	set_info (L);
